@@ -104,7 +104,7 @@ let rec like_i64 t =
 let follow_once t =
 	match t with
 	| TMono r ->
-		(match !r with
+		(match r.tm_type with
 		| Some t -> t
 		| _ -> t_dynamic) (* avoid infinite loop / should be the same in this context *)
 	| TLazy f ->
@@ -116,7 +116,7 @@ let follow_once t =
 	| _ ->
 		t
 
-let t_empty = TAnon({ a_fields = PMap.empty; a_status = ref Closed })
+let t_empty = mk_anon (ref Closed)
 
 let alloc_var n t = Type.alloc_var VGenerated n t null_pos
 
@@ -163,10 +163,10 @@ let anon_class t =
 			| EnumStatics e -> TEnumDecl e
 			| Statics cl -> TClassDecl cl
 			| AbstractStatics a -> TAbstractDecl a
-			| _ -> assert false)
+			| _ -> die "" __LOC__)
 	| TLazy f -> t_to_md (lazy_type f)
-	| TMono r -> (match !r with | Some t -> t_to_md t | None -> assert false)
-	| _ -> assert false
+	| TMono r -> (match r.tm_type with | Some t -> t_to_md t | None -> die "" __LOC__)
+	| _ -> die "" __LOC__
 
 
 let get_cl mt = match mt with TClassDecl cl -> cl | _ -> failwith (Printf.sprintf "Unexpected module type (class expected) for %s: %s" (s_type_path (t_path mt)) (s_module_type_kind mt))
@@ -175,7 +175,7 @@ let get_abstract mt = match mt with TAbstractDecl a -> a | _ -> failwith (Printf
 let get_fun t =
 	match follow t with
 	| TFun (args, ret) -> args, ret
-	| t -> (trace (debug_type t)); assert false
+	| t -> (trace (debug_type t)); die "" __LOC__
 
 let mk_cast t e = Type.mk_cast e t e.epos
 
@@ -378,6 +378,8 @@ type generator_ctx =
 	(* this is all you need to care about *)
 	gcon : Common.context;
 
+	gentry_point : (string * tclass * texpr) option;
+
 	gclasses : gen_classes;
 
 	gtools : gen_tools;
@@ -566,19 +568,20 @@ let new_ctx con =
 		| TClassDecl c -> c
 		| TAbstractDecl a ->
 				mk_class a.a_module ([], "Dynamic") a.a_pos null_pos
-		| _ -> assert false
+		| _ -> die "" __LOC__
 	in
 
 	let rec gen = {
 		gcon = con;
+		gentry_point = get_entry_point con;
 		gclasses = {
 			cl_reflect = get_cl (get_type ([], "Reflect"));
 			cl_type = get_cl (get_type ([], "Type"));
 			cl_dyn = cl_dyn;
 
-			nativearray = (fun _ -> assert false);
-			nativearray_type = (fun _ -> assert false);
-			nativearray_len = (fun _ -> assert false);
+			nativearray = (fun _ -> die "" __LOC__);
+			nativearray_type = (fun _ -> die "" __LOC__);
+			nativearray_len = (fun _ -> die "" __LOC__);
 		};
 		gtools = {
 			r_fields = (fun is_used_only_by_iteration expr ->
@@ -597,7 +600,7 @@ let new_ctx con =
 				mk_cast t { eexpr = TCall(fieldcall, [obj; field]); etype = t_dynamic; epos = obj.epos }
 			);
 
-			r_create_empty = (fun _ _ pos -> gen.gcon.error "r_create_empty implementation is not provided" pos; assert false);
+			r_create_empty = (fun _ _ pos -> gen.gcon.error "r_create_empty implementation is not provided" pos; die "" __LOC__);
 		};
 		gexpr_filters = new rule_map_dispatcher "gexpr_filters";
 		gmodule_filters = new rule_map_dispatcher "gmodule_filters";
@@ -656,7 +659,7 @@ let init_ctx gen =
 	let follow t =
 		match t with
 		| TMono r ->
-			(match !r with
+			(match r.tm_type with
 			| Some t -> follow_f t
 			| _ -> Some t)
 		| TLazy f ->
@@ -836,7 +839,7 @@ let write_file gen w source_dir path extension out_files =
 		close_out f
 	end;
 
-	out_files := (Path.unique_full_path s_path) :: !out_files;
+	out_files := (Path.UniqueKey.create s_path) :: !out_files;
 
 	t()
 
@@ -851,7 +854,7 @@ let clean_files path excludes verbose =
 				let pack = pack @ [file] in
 				iter_files (pack) (Unix.opendir filepath) filepath;
 				try Unix.rmdir filepath with Unix.Unix_error (ENOTEMPTY,_,_) -> ();
-			else if not (String.ends_with filepath ".meta") && not (List.mem (Path.unique_full_path filepath) excludes) then begin
+			else if not (String.ends_with filepath ".meta") && not (List.mem (Path.UniqueKey.create filepath) excludes) then begin
 				if verbose then print_endline ("Removing " ^ filepath);
 			 	Sys.remove filepath
 			end
@@ -915,19 +918,19 @@ let dump_descriptor gen name path_s module_s =
 	SourceWriter.write w "end modules";
 	SourceWriter.newline w;
 	(* dump all resources *)
-	(match gen.gcon.main_class with
-		| Some path ->
-			SourceWriter.write w "begin main";
-			SourceWriter.newline w;
-			(try
-				SourceWriter.write w (Hashtbl.find main_paths path)
-			with
-				| Not_found -> SourceWriter.write w (path_s path));
-			SourceWriter.newline w;
-			SourceWriter.write w "end main";
-			SourceWriter.newline w
-	| _ -> ()
-	);
+	(match gen.gentry_point with
+	| Some (_,cl,_) ->
+		SourceWriter.write w "begin main";
+		SourceWriter.newline w;
+		let path = cl.cl_path in
+		(try
+			SourceWriter.write w (Hashtbl.find main_paths path)
+		with Not_found ->
+			SourceWriter.write w (path_s path));
+		SourceWriter.newline w;
+		SourceWriter.write w "end main";
+		SourceWriter.newline w
+	| _ -> ());
 	SourceWriter.write w "begin resources";
 	SourceWriter.newline w;
 	Hashtbl.iter (fun name _ ->
@@ -1010,7 +1013,7 @@ let follow_module follow_func md = match md with
 		| TEnum(e,_) -> TEnumDecl e
 		| TType(t,_) -> TTypeDecl t
 		| TAbstract(a,_) -> TAbstractDecl a
-		| _ -> assert false
+		| _ -> die "" __LOC__
 
 (*
 	hxgen means if the type was generated by haxe. If a type was generated by haxe, it means
@@ -1045,14 +1048,14 @@ let mt_to_t mt params =
 		| TClassDecl (cl) -> TInst(cl, params)
 		| TEnumDecl (e) -> TEnum(e, params)
 		| TAbstractDecl a -> TAbstract(a, params)
-		| _ -> assert false
+		| _ -> die "" __LOC__
 
 let t_to_mt t =
 	match follow t with
 		| TInst(cl, _) -> TClassDecl(cl)
 		| TEnum(e, _) -> TEnumDecl(e)
 		| TAbstract(a, _) -> TAbstractDecl a
-		| _ -> assert false
+		| _ -> die "" __LOC__
 
 let rec get_last_ctor cl =
 	Option.map_default (fun (super,_) -> if is_some super.cl_constructor then Some(get super.cl_constructor) else get_last_ctor super) None cl.cl_super
@@ -1068,8 +1071,8 @@ let add_constructor cl cf =
 let rec replace_mono t =
 	match t with
 	| TMono t ->
-		(match !t with
-		| None -> t := Some t_dynamic
+		(match t.tm_type with
+		| None -> Monomorph.bind t t_dynamic
 		| Some _ -> ())
 	| TEnum (_,p) | TInst (_,p) | TType (_,p) | TAbstract (_,p) ->
 		List.iter replace_mono p
@@ -1082,8 +1085,8 @@ let rec replace_mono t =
 		replace_mono (lazy_type f)
 
 (* helper *)
-let mk_class_field name t public pos kind params =
-	let f = mk_field name ~public t pos null_pos in
+let mk_class_field ?(static = false) name t public pos kind params =
+	let f = mk_field name ~public ~static t pos null_pos in
 	f.cf_meta <- [ Meta.CompilerGenerated, [], null_pos ]; (* annotate that this class field was generated by the compiler *)
 	f.cf_kind <- kind;
 	f.cf_params <- params;
@@ -1099,7 +1102,7 @@ let map_param cl =
 	ret
 
 let get_cl_t t =
-	match follow t with | TInst (cl,_) -> cl | _ -> assert false
+	match follow t with | TInst (cl,_) -> cl | _ -> die "" __LOC__
 
 let mk_class m path pos =
 	let cl = Type.mk_class m path pos null_pos in
@@ -1223,7 +1226,7 @@ let rec field_access gen (t:t) (field:string) : (tfield_access) =
 			   but for now, we're going to find the generated class and make a field access to it instead. *)
 			(try
 				let cl_enum = List.find (function TClassDecl cl when cl.cl_path = en.e_path && Meta.has Meta.Enum cl.cl_meta -> true | _ -> false) gen.gtypes_list in
-				let cl_enum = match cl_enum with TClassDecl cl -> TInst (cl,params) | _ -> assert false in
+				let cl_enum = match cl_enum with TClassDecl cl -> TInst (cl,params) | _ -> die "" __LOC__ in
 				field_access gen cl_enum field
 			with Not_found ->
 				FNotFound)
@@ -1274,7 +1277,7 @@ let mk_field_access gen expr field pos =
 				{ eexpr = TField(expr, FDynamic field); etype = t; epos = pos }
 		| FNotFound ->
 				{ eexpr = TField(expr, FDynamic field); etype = t_dynamic; epos = pos }
-		| FEnumField _ -> assert false
+		| FEnumField _ -> die "" __LOC__
 
 (* ******************************************* *)
 (* Module dependency resolution *)
